/*
  Copyright 2021 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.partner.friend.view

import android.animation.LayoutTransition
import android.app.Activity
import android.content.res.Configuration
import android.os.Bundle
import android.os.Parcelable
import android.text.Editable
import android.text.TextWatcher
import android.util.TypedValue
import android.view.Gravity
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.LinearLayout
import android.widget.Toast
import androidx.constraintlayout.widget.ConstraintSet
import androidx.constraintlayout.widget.ConstraintSet.*
import androidx.core.content.res.ResourcesCompat
import androidx.core.graphics.drawable.DrawableCompat
import androidx.fragment.app.Fragment
import androidx.lifecycle.Observer
import androidx.lifecycle.ViewModelProvider
import androidx.recyclerview.widget.SimpleItemAnimator
import com.kakao.sdk.common.model.ClientError
import com.kakao.sdk.common.model.ClientErrorCause
import com.kakao.sdk.common.model.KakaoSdkError
import com.kakao.sdk.partner.friend.Constants
import com.kakao.sdk.partner.friend.adapter.FriendsListAdapter
import com.kakao.sdk.partner.friend.adapter.PickedFriendsAdapter
import com.kakao.sdk.partner.friend.model.*
import com.kakao.sdk.partner.friend.util.LayoutUtils
import com.kakao.sdk.partner.friend.util.dpToPixel
import com.kakao.sdk.partner.friend.viewmodel.FriendViewModel
import com.kakao.sdk.v2.partner.friend.R
import com.kakao.sdk.v2.partner.friend.databinding.*
import kotlinx.android.synthetic.main.friends_list.*

/**
 * @suppress
 */
internal class FriendPickerFragment : Fragment() {
    private var _binding: FragmentFriendPickerBinding? = null
    private val binding get() = _binding!!

    private var title = ""
    private var friendType = PickerFriendFilter.NONE
    private var isSingle = false // option
    private var isPopup = false // option
    private var enableSearch = true // option
    private var enableIndex = true // option
    private var searchPlaceHolder = "" // option
    private var showPickedFriend = true // option
    private var showMyProfile = true // option
    private var showFavorite = true // option
    private var maxPickableCount = 1
    private var minPickableCount = 1

    private var titleBar: TitleBarBinding? = null
    private var pickedFriendsRv: PickedFriendsViewBinding? = null
    private var searchBarLayout: SearchBarBinding? = null
    private var friendsRv: FriendsListBinding? = null

    private val pickedFriendsAdapter: PickedFriendsAdapter by lazy {
        PickedFriendsAdapter(isPopup, viewModel.pickedFriendsList) {
            friendListAdapter.uncheckFriend(it)
            viewModel.removeUncheckedFriend(it)
        }
    }

    private val friendListAdapter: FriendsListAdapter by lazy {
        FriendsListAdapter(
            isPopup, isSingle, maxPickableCount,
            checkCallback = {
                // 선택 가능한 숫자 이상으로 친구를 선택한 경우 null 리턴 받음
                if (it == null) {
                    Toast.makeText(
                        context,
                        getString(R.string.max_pickable_count_message, maxPickableCount),
                        Toast.LENGTH_SHORT
                    ).show()
                } else {
                    pickedFriendsAdapter.addPickedFriend(it)
                    pickedFriendsRv?.pickedFriendsRv?.scrollToPosition(0)
                    friendListAdapter.checkFriend(it)
                    viewModel.addCheckedFriend(it)
                }
            },
            uncheckCallback = {
                pickedFriendsAdapter.removePickedFriend(it)
                viewModel.removeUncheckedFriend(it)
                friendListAdapter.uncheckFriend(it)
            }
        )
    }
    private val viewModel by lazy {
        ViewModelProvider(requireActivity(), ViewModelProvider.NewInstanceFactory()).get(
            FriendViewModel::class.java
        )
    }

    override fun onCreateView(
        inflater: LayoutInflater, container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View {
        _binding = FragmentFriendPickerBinding.inflate(layoutInflater, container, false)
        return binding.root
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)
        setProperty()
        createLayout()

        // 최초 뷰 생성시 savedInstanceState가 null이므로 화면 회전 구분
        if (savedInstanceState == null) {
            viewModel.fetchData() // 최초 뷰 생성시에만 서버로부터 데이터 로딩
        }
        initializeView()
        observeData()
    }

    private fun createLayout() {
        binding.mainLayout.apply {
            titleBar = TitleBarBinding.inflate(layoutInflater, this, false).apply {
                title.text = this@FriendPickerFragment.title
                if (isSingle) {
                    pickedCountTv.visibility = View.GONE
                }
                addView(root)
            }
            if (!isSingle && showPickedFriend) {
                pickedFriendsRv =
                    PickedFriendsViewBinding.inflate(layoutInflater, this, false).apply {
                        addView(root)
                    }
            }
            if (enableSearch) {
                searchBarLayout = SearchBarBinding.inflate(layoutInflater, this, false).apply {
                    searchBar.hint = searchPlaceHolder
                    addView(root)
                }
                searchBarLayout?.apply {
                    searchBar.addTextChangedListener(object : TextWatcher {
                        override fun beforeTextChanged(
                            s: CharSequence?,
                            start: Int,
                            count: Int,
                            after: Int
                        ) {
                        }

                        override fun onTextChanged(
                            s: CharSequence?,
                            start: Int,
                            before: Int,
                            count: Int
                        ) {
                        }

                        override fun afterTextChanged(s: Editable?) {
                            searchBarDeleteBtn.visibility =
                                if (s.isNullOrEmpty()) View.GONE else View.VISIBLE
                            viewModel.searchFriends(s.toString())
                        }
                    })
                    searchBarDeleteBtn.setOnClickListener {
                        searchBar.setText("")
                    }
                }
            }
            friendsRv = FriendsListBinding.inflate(layoutInflater, this, false).apply {
                addView(root, LinearLayout.LayoutParams(ViewGroup.LayoutParams.MATCH_PARENT, 0, 1f))

                if (enableIndex) {
                    sideIv.recyclerView = friendsList
                    val wrapped = DrawableCompat.wrap(sideIv.background)
                    val typedValue = TypedValue()
                    context.theme.resolveAttribute(R.attr.gray050s, typedValue, true)
                    DrawableCompat.setTint(wrapped.mutate(), typedValue.data)
                    friendsList.addOnScrollListener(sideIv.onScrollListener)
                }
            }
        }

        if (LayoutUtils.orientation != Configuration.ORIENTATION_PORTRAIT) {
            binding.emptyFriendView.emptyFriendImage.visibility = View.GONE
        }

        if (isPopup) {
            setPopupLayout()
        }
    }

    private fun initializeView() {
        pickedFriendsRv?.pickedFriendsRv?.adapter = pickedFriendsAdapter
        friendsRv?.friendsList?.apply {
            adapter = friendListAdapter
            // 즐겨찾기, 친구 리스트에 중복되는 친구를 터치할 경우 깜빡거리는 현상 제거
            itemAnimator.apply {
                if (this is SimpleItemAnimator) {
                    supportsChangeAnimations = false
                }
            }
        }
        titleBar?.apply {
            backBtn.setOnClickListener {
                sendError(ClientError(ClientErrorCause.Cancelled, "cancelled."))
            }
            doneBtn.setOnClickListener {
                sendPickedFriends()
            }
        }
    }

    private fun setPopupLayout() {
        titleBar?.apply {
            if (LayoutUtils.orientation == Configuration.ORIENTATION_PORTRAIT) {
                root.setPadding(0, dpToPixel(8f, resources), 0, 0)
            } else {
                root.setPadding(0, dpToPixel(4f, resources), 0, 0)
            }
            backBtn.apply {
                setImageDrawable(
                    ResourcesCompat.getDrawable(
                        resources,
                        R.drawable.friendspicker_btn_close,
                        null
                    )
                )
                ConstraintSet().apply {
                    clone(root)
                    connect(id, START, PARENT_ID, START, dpToPixel(14f, resources))
                    connect(id, TOP, PARENT_ID, TOP, dpToPixel(7f, resources))
                    connect(id, BOTTOM, PARENT_ID, BOTTOM, dpToPixel(7f, resources))
                }.applyTo(root)
            }
            title.apply {
                gravity = Gravity.CENTER
                setTextSize(TypedValue.COMPLEX_UNIT_SP, 15f)
                ConstraintSet().apply {
                    clone(root)
                    connect(id, START, PARENT_ID, START)
                    connect(id, TOP, PARENT_ID, TOP, dpToPixel(7f, resources))
                    connect(id, END, PARENT_ID, END)
                    connect(id, BOTTOM, PARENT_ID, BOTTOM, dpToPixel(8.5f, resources))
                }.applyTo(root)
            }
            doneBtn.apply {
                ConstraintSet().apply {
                    clone(doneBtnLayout)
                    connect(id, TOP, PARENT_ID, TOP, dpToPixel(7f, resources))
                    connect(id, BOTTOM, PARENT_ID, BOTTOM, dpToPixel(8.5f, resources))
                }.applyTo(doneBtnLayout)
            }
        }
        searchBarLayout?.apply {
            searchBarBackground.apply {
                ConstraintSet().apply {
                    clone(root)
                    connect(id, TOP, PARENT_ID, TOP, dpToPixel(6f, resources))
                }.applyTo(root)
            }
            searchBar.setTextSize(TypedValue.COMPLEX_UNIT_SP, 13f)
        }
        friendsRv?.friendsList?.apply {
            if (LayoutUtils.orientation == Configuration.ORIENTATION_PORTRAIT) {
                setPadding(0, 0, 0, dpToPixel(8f, resources))
            } else {
                setPadding(0, 0, 0, dpToPixel(4f, resources))
            }
        }
    }

    private fun observeData() {
        viewModel.run {
            pickerList.observe(viewLifecycleOwner, Observer {
                friendListAdapter.setFriendsList(it)
                friendsRv?.friendsList?.scrollToPosition(0)

                if (enableIndex) {
                    val friends = viewModel.friendsList
                    var friendListOffset = 0
                    var favoriteOffset = -1

                    it.forEach { item ->
                        ++friendListOffset
                        if (item is Item.Header && item.title == getString(R.string.favorite)) {
                            favoriteOffset = friendListOffset - 1
                            return@forEach
                        }
                    }

                    // extraOffset = RecyclerView 전체 리스트 - 순수 친구 리스트 - favoriteOffset
                    val extraOffset = it.size - friends.size - favoriteOffset

                    sideIv.setDataSource(friends, friends.size, favoriteOffset, extraOffset)
                    // 팝업 유무에 따라서 인덱스 리스트 변경
                    if (isPopup) {
                        sideIv.setIndexItems(
                            R.array.popup_side_indexer,
                            R.array.popup_side_indexer_landscape
                        )
                    } else {
                        sideIv.setIndexItems(R.array.side_indexer, R.array.side_indexer_landscape)
                    }
                }
            })
            pickedFriends.observe(viewLifecycleOwner, Observer {
                friendListAdapter.setPickedFriends(it)
                pickedFriendsRv?.root?.visibility = if (it.size > 0) View.VISIBLE else View.GONE
                titleBar?.pickedCountTv?.text = if (it.size > 0) it.size.toString() else ""
                titleBar?.doneBtn?.isEnabled = it.size > 0
            })
            isSearching.observe(viewLifecycleOwner, Observer {
                if (it) {
                    friendsRv?.friendsList?.removeOnScrollListener(sideIv.onScrollListener)
                } else {
                    friendsRv?.friendsList?.addOnScrollListener(sideIv.onScrollListener)
                }
            })
            isError.observe(viewLifecycleOwner, Observer {
                if (it) {
                    binding.root.layoutTransition = null // 검색창 사라질 때 애니메이션 효과 제거
                    titleBar?.doneBtn?.visibility = View.GONE
                    searchBarLayout?.root?.visibility = View.GONE
                    binding.emptyFriendView.root.visibility = View.VISIBLE
                } else {
                    binding.root.layoutTransition = LayoutTransition()
                    titleBar?.doneBtn?.visibility = View.VISIBLE
                    searchBarLayout?.root?.visibility = View.VISIBLE
                    binding.emptyFriendView.root.visibility = View.GONE
                }
            })
        }
    }

    private fun setProperty() {
        viewModel.param.also {
            if (it is PickerRequestParams) {
                title = it.title
                friendType = it.pickerFriendFilter!!
                enableSearch = it.enableSearch!!
                enableIndex = it.enableIndex!!
                searchPlaceHolder = it.searchPlaceHolder ?: getString(R.string.search_bar_hint)
                showMyProfile = it.showMyProfile!!
                showFavorite = it.showFavorite!!
            }
            if (it is MultiPickerRequestParams) {
                minPickableCount = it.minPickableCount!!
                maxPickableCount = it.maxPickableCount!!
                showPickedFriend = it.showPickedFriend!!
            }
        }
        isSingle = viewModel.isSingle
        isPopup = viewModel.isPopup
    }

    override fun onDestroyView() {
        super.onDestroyView()
        _binding = null
    }

    private fun sendPickedFriends() {
        if (viewModel.pickedFriendsList.size < minPickableCount) {
            Toast.makeText(
                requireContext(),
                getString(R.string.min_pickable_count_message, minPickableCount),
                Toast.LENGTH_LONG
            ).show()
        } else {
            sendOK()
        }
    }

    private fun sendOK() {
        val friends = mutableListOf<SelectedUser>()
        viewModel.pickedFriendsList.forEach {
            friends.add(SelectedUser(it.nickname, it.uuid))
        }
        viewModel.resultReceiver.send(
            Activity.RESULT_OK,
            Bundle().apply {
                putParcelableArrayList(
                    Constants.KEY_SELECTED_FRIENDS,
                    ArrayList<Parcelable>(friends)
                )
            }
        )
        requireActivity().finish()
    }

    private fun sendError(exception: KakaoSdkError) {
        viewModel.resultReceiver.send(
            Activity.RESULT_CANCELED,
            Bundle().apply { putSerializable(Constants.KEY_EXCEPTION, exception) }
        )
        requireActivity().finish()
    }
}
